import io
import sys
import traceback

from .BaseTest import BaseTest
from .TestResponse import TestResponse
from autograder.printing import StatusMessage
from autograder.io_utils import RedirectStdin

class ArgTest(BaseTest):
    def __init__(self,
                 student_obj=None,
                 solution_obj=None,
                 args=(),
                 kwargs={},
                 start_msg=None,
                 *pos_args,
                 **key_args):
        """
        Initializes the BaseTest object which compares the student object to
        the solution object.

        Arguments
        ---------
        args (tuple) -- The positional arguments to supply to the function when 
            tested.
        kwargs (dict) -- The keyword arguments to supply to the function when
            tested.
        start_msg (str or None) -- The message to print at the beginning of the 
            test.
        """
        super().__init__(student_obj, solution_obj, start_msg,
                         *pos_args, **key_args)

        self.args = args
        self.kwargs = kwargs

        # Optionally overwrite the start message with the filled out arguments
        if start_msg is None:
            self.start_msg = f"Testing {self._serialize_args()}..."
            self.start_msg = "{:68}".format(self.start_msg)


    def _serialize_args(self):
        """
        Builds a string representing the function call.
        """
        student_obj_name = self.student_obj.__name__
        
        args_lst = [repr(arg) for arg in self.args]
        kwargs_lst = [f'{k}={repr(v)}' for k, v in self.kwargs.items()]
        all_args = args_lst + kwargs_lst

        output = f"{student_obj_name}({', '.join(all_args)})"

        if len(output) > 57:
            output = f"{student_obj_name}(...several arguments...)"

        return output


    @staticmethod
    def _captured_runner(fn, args, kwargs, name,
                         f_stdout=None, f_stderr=None, f_stdin=None):
        """
        Runs fn with args, kwargs and loads responses into a TestResponse object
        with name.

        Keyword Arguments
        -----------------
        f_stdout, f_stderr, f_stdin (buffer or None) -- The buffer to read/write
            the captured data from/to.
        """
        # Store old buffers
        __old_stdout = sys.stdout
        __old_stderr = sys.stderr
        __old_stdin = sys.stdin

        # Create new buffers
        f_stdout = f_stdout or io.StringIO()
        f_stderr = f_stderr or io.StringIO()
        f_stdin = f_stdin or RedirectStdin()

        # Replace buffers
        sys.stdout = f_stdout
        sys.stderr = f_stderr
        sys.stdin = f_stdin

        # Build TestRunner arguments
        output = None
        error = None
        warning = None

        # Test the function
        try:
            output = fn(*args, **kwargs)

        except Exception as e:
            # Function raised an exception
            error = f"Threw {e}.\n{traceback.format_exc()}"

        except SystemExit:
            # Function called exit() or quit()
            warning = 'The code tried to exit the Python process.'

        finally:
            # Restore the old buffers
            sys.stdout = __old_stdout
            sys.stderr = __old_stderr
            sys.stdin = __old_stdin

        # If there was an uncaught exception above, the below code will not run
        stdout = f_stdout.getvalue()
        stderr = f_stderr.getvalue()

        return TestResponse(output, stdout, stderr, name, error, warning)


    def run(self):
        """
        Runs the argument test (captures stdout and stderr and runs both of the
        functions).
        """
        self._setup()

        self.solution_response = self._captured_runner(
            self.solution_obj, self.args, self.kwargs, 'solution'
        )

        self.student_response = self._captured_runner(
            self.student_obj, self.args, self.kwargs, 'student'
        )

        return self._process_responses()


import collections
from difflib import unified_diff

from autograder.printing import StatusMessage

BaseTestResponse = collections.namedtuple(
    'TestResponse',
    ('output', 'stdout', 'stderr', 'name', 'error', 'warning'),
    defaults=(None, None)
)

class TestResponse(BaseTestResponse):
    def __init__(self, *args, **kwargs):
        if self.stderr and not self.warning:
            self.warning = self.stderr

        self.has_output_diff = False


    def __eq__(self, other):
        """
        Compares two TestResponse objects. Checks that the return value was the 
        same, the printed output was the same, and there was no difference in
        error throwing.
        """
        if self.error and other.error:
            # Don't care about anything else
            return True

        return (self.output == other.output \
                and self.stdout == other.stdout \
                and bool(self.error) == bool(other.error))


    def diff(self, other, self_name=None, other_name=None):
        """
        Returns a diff between self and another test response object. This is
        intended for comparing the student output to the solution output.

        Arguments
        ---------
        other (TestResponse) -- The test response object to compare to.
        self_name (str or None) -- The title of the current test response 
            object.
        other_name (str or None) -- The title of the other test response object.

        Returns
        -------
        str -- The diff between self and other.
        """
        self_name = self_name or self.name
        other_name = other_name or other.name

        # Case 1: An error occurred during the execution of the code.
        error_occurred = self.error or other.error

        if error_occurred:
            both_errored = self.error and other.error
            if not both_errored:
                # Problem: one raised an error and the other didn't.
                if self.error:
                    header = f"{self_name.title()} threw an unexpected error:"
                    error_line = f"\n{self.error}"
                else:
                    header = f"{other_name.title()} threw an unexpected error:"
                    error_line = f"\n{other.error}"

                return (
                    f"{StatusMessage(header, 'info')}"
                    f"{error_line}"
                )

            return

        # Case 2: Difference in what was printed.
        printing_error = self.stdout != other.stdout

        if printing_error:
            self_output = self.stdout.splitlines(keepends=True)
            other_output = other.stdout.splitlines(keepends=True)
            diff = ''.join(unified_diff(
                self_output, 
                other_output, 
                fromfile=self_name, 
                tofile=other_name
            )).strip()

            self.has_output_diff = True # Michael's thingy

            return (f"{StatusMessage('Difference in printed output:', 'info')}"
                    f"\n"
                    f"{diff}")

        # Case 3: Difference in return value.
        return_error = self.output != other.output

        if return_error:
            return (
                f"{StatusMessage('Difference in value:', 'info')}\n"
                f"{self_name}: {repr(self.output)}\n"
                f"{other_name}: {repr(other.output)}"
            )

        # Case 4: Warning!
        warning = self.warning or other.warning

        if warning:
            both_warned = self.warning and other.warning
            if not both_warned:
                # Problem: one raised a warning.
                if self.error:
                    header = f"{self_name} caused a warning:"
                    warning_line = f"{self.warning}"
                else:
                    header = f"{other_name} caused a warning:"
                    warning_line = f"{other.warning}"

                return (
                    f"{StatusMessage(header, 'warning')} "
                    f"{warning_line}"
                )

        # Otherwise: No issues!


import contextlib
import io
import multiprocessing as mp
import sys
from autograder.tests import BaseTest
from .printing import StatusMessage

class TestRunner:
    def __init__(self, res_queue, print_val, print_condn):
        """
        A pickle-able object to use for multiprocessing test running.

        Arguments
        ---------
        res_queue (mp.Queue) -- The queue in which to add the result of the 
            test.
        print_val (mp.Value) -- The value of the thread that has the okay to 
            print.
        print_condn (mp.Condition) -- The condition on which the process should
            wait to print
        """
        self.queue = res_queue
        self.print_val = print_val
        self.print_condn = print_condn


    def __call__(self, index, test):
        # Run test
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            passed = test.run()
        out = f.getvalue()

        # Queue result and test number
        self.queue.put((index, passed))

        # Print result
        with self.print_condn:
            while not self.print_val.value == index:
                self.print_condn.wait()
        
            print(out, end='')
            self.print_val.value += 1
            self.print_condn.notify_all()


class TestSuite:
    def __init__(self, tests=[], multiprocess=False, ml=None):
        """
        A collection of tests to be run together. Supports multiprocessing and
        ML integration.

        ML Integration:
            ml should be a function which accepts a list of 1s and 0s. That list
            will signify the tests that the program passes (1) and fails (0) in
            the correct order.
        """
        # Initialize the tests
        self.tests = []
        for test in tests:
            self.add_test(test)

        self.multiprocess = multiprocess
        self.ml = ml


    def add_test(self, test):
        """
        Adds test to the internal collection of tests.
        """
        if not isinstance(test, BaseTest):
            raise ValueError(
                f"Can't write a {type(test)} object to the TestRunner (expected"
                f" a test object)."
            )

        self.tests.append(test)


    def _close_suite(self, num_tests, num_passed):
        status = 'success' if num_tests == num_passed else 'warning'

        print()
        print(StatusMessage(
            f"{num_passed} / {num_tests} tests passed.",
            status
        ))

        if self.ml:
            # Hand the test information to the ML model
            print()
            self.ml(self.pass_list)


    def _run_mp(self):
        """
        Runs the tests in a multiprocessing pool.
        """
        # Twice as many workers because why not?
        num_workers = mp.cpu_count() * 2
        manager = mp.Manager()
        passed_q = manager.Queue()

        # Two resources for printing: the process that can print and the condn
        print_val = manager.Value(int, 0)
        print_condn = manager.Condition()

        # Put the tests into a pool.
        with mp.Pool(num_workers) as p:
            p.starmap(
                TestRunner(passed_q, print_val, print_condn), 
                enumerate(self.tests)
            )

        # Calculate the number that passed and build a list for ML
        num_passed = 0
        self.pass_list = [0] * len(self.tests)
        while not passed_q.empty():
            index, passed = passed_q.get()
            if passed:
                num_passed += 1
                self.pass_list[index] = 1

        self._close_suite(len(self.tests), num_passed)


    def _run_normal(self):
        """
        Runs all of the tests in order.
        """
        num_passed, num_tests = 0, len(self.tests)
        self.pass_list = []

        for test in self.tests:
            if test.run():
                # Test passed
                num_passed += 1
                self.pass_list.append(1)
            else:
                self.pass_list.append(0)

        self._close_suite(num_tests, num_passed)


    def run(self):
        # Progressive mode cannot run with multiprocessing.
        is_progressive = '-p' in sys.argv or '--progressive' in sys.argv

        if self.multiprocess and (not is_progressive):
            self._run_mp()

        else:
            if self.multiprocess:
                no_progressive = StatusMessage(
                    ("Progressive mode is incompatible with multiprocessing. "
                     "The autograder will run \nin a single process."),
                    'warning'
                )
                print(no_progressive)
            self._run_normal()


import sys
import io

class StatusMessage:
    """
    A class designed to create colorful messages to output during the
    execution of the autograder. This object is designed to be printed
    immediately.
    """

    def __init__(self, message, status=None):
        """
        Usage:
            print(StatusMessage("Passed all tests!", "success"))

        status -- The available statuses are
                'success'
                'fail'
                'warning'
                'info'
                'bold'
                'underline'
        """
        self._setup_color_support()

        self._plain_status = status
        self._plain_message = message
        self.message = self._get_color(message, status)


    def __str__(self):
        return self.message


    def __repr__(self):
        color_flag = self._color if self._colorized else 'NO_COLOR_SUPPORT'

        return (f'<StatusMessage: "{self._plain_message}", {self._plain_status}'
                f' / {color_flag}>')


    def _supports_color(self):
        """
        Returns True if the terminal supports color printing, False
        otherwise. Creds to Stack Overflow.
        """
        plat = sys.platform
        supported_platform = plat != 'Pocket PC' and (
            plat != 'win32' or 'ANSICON' in os.environ)

        # isatty is not always implemented, #6223.
        is_a_tty = (hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()) \
                   or isinstance(sys.stdout, io.IOBase)

        if not supported_platform or not is_a_tty:
            return False

        return True


    def _setup_color_support(self):
        """
        Sets up the class based on whether the terminal suppports
        colorized printing and which modules can be used to do that.
        """
        # color support?
        self._colorized = self._supports_color()

        # termcolor?
        try:
            from termcolor import colored
        except ModuleNotFoundError:
            # the module does not exist
            self._use_termcolor = False
        else:
            self._use_termcolor = True


    def _get_color(self, message, status):
        """
        Converts the status to a color hex
        """
        # set up color / status maps
        COLOR_HEX = {
            'header': '\033[95m',
            'blue': '\033[94m',
            'green': '\033[92m',
            'yellow': '\033[93m',
            'red': '\033[91m',
            'end': '\033[0m',
            'bold': '\033[1m',
            'underline': '\033[4m',
        }

        STATUS_COLORS = {
            'success': 'green',
            'fail': 'red',
            'warning': 'yellow',
            'info': 'blue',
            'bold': 'bold',
            'underline': 'underline',
        }

        # Neutral status option
        if status is None:
            return message

        # colorize the text
        if self._colorized:
            try:
                self._color = STATUS_COLORS[status]
            except KeyError:
                raise KeyError(f"'{status}' is not a valid status.") from None

            if self._use_termcolor:
                # termcolor... way more reliable than the hack below
                from termcolor import colored
                return colored(message, self._color)

            else:
                # colored ascii printing
                return COLOR_HEX[self._color] + message + COLOR_HEX['end']
        else:
            # plaintext
            return message

class HeaderMessage(StatusMessage):
    """
    Extends StatusMessage with a border around the message
    """

    def __init__(self, message, status=None):
        horiz_lines = max(len(message) + 10, 80)

        # new_message = [
        #     '┌' + '─' * horiz_lines + '┐',
        #     '│ {:^{spacing}} │'.format(message, spacing=horiz_lines-2),
        #     '└' + '─' * horiz_lines + '┘',
        # ]
        # new_message = '\n'.join(new_message)

        new_message = (
            "{:^{spacing}}\n".format(message, spacing=horiz_lines)
            + '–' * horiz_lines
        )

        super().__init__(new_message, status)

class SuperHeaderMessage(StatusMessage):
    """
    Extends StatusMessage with a thicc border around the message
    """

    def __init__(self, message, status=None):
        horiz_lines = max(len(message) + 10, 78)

        # new_message = [
        #     '╔' + '═' * horiz_lines + '╗',
        #     '║ {:^{spacing}} ║'.format(message, spacing=horiz_lines-2),
        #     '╚' + '═' * horiz_lines + '╝',
        # ]
        # new_message = '\n'.join(new_message)

        new_message = [
            '┌' + '─' * horiz_lines + '┐',
            '│ {:^{spacing}} │'.format(message, spacing=horiz_lines-2),
            '└' + '─' * horiz_lines + '┘',
        ]
        new_message = '\n'.join(new_message)

        super().__init__(new_message, status)



__version__ = '0.2'
__author__ = 'Parth Sarin'
__license__ = 'MIT'

import py_compile
import pycodestyle
import contextlib
import io

from .printing import StatusMessage, HeaderMessage, SuperHeaderMessage

class Autograder:
    """
    The basic autograder class.
    """

    def __init__(self,
                 module_name=None,
                 module_overrides={},
                 has_compile_check=True,
                 has_custom_tests=False,
                 has_style_tests=True):
        """
        Initializes the autograder.

        Arguments
        ---------
        module_name (str or None) -- The name of the student module.
        module_overrides (dict) -- Namespace overrides for the module.
        has_compile_check (bool) -- Whether to run a basic compile check.
        has_custom_tests (bool) -- Whether the autograder has a functioning
            run_custom_tests function.
        has_style_tests (bool) -- Whether to run PEP8 style checking on the
            module.
        """
        self.module_name = module_name.rstrip('.py')
        self.module_overrides = module_overrides

        self.has_compile_check = has_compile_check
        self.has_custom_tests = has_custom_tests
        self.has_style_tests = has_style_tests


    def run(self):    
        """
        Run the autograder.
        """
        print(SuperHeaderMessage(
            f"Starting Autograder for {self.module_name}...",
            "info"
        ))

        self._load_module()

        # Overwrites on the module
        self.module.__dict__.update(self.module_overrides)

        if self.has_custom_tests:
            print(HeaderMessage(
                "Running custom tests on {}...".format(self.module_name)
            ))
            self.run_custom_tests()

        if self.has_style_tests:
            self.run_style_tests()


    def _load_module(self):
        """
        Loads the module by either checking compile or directly importing,
        based on the preferences given in __init__.
        """
        if self.has_compile_check:
            self.run_compile_check() # will import the module
        else:
            self.module = __import__(self.module_name)

        return self.module


    def run_compile_check(self):
        """
        Checks that the module does not have syntax errors.
        """
        print(HeaderMessage(
            "Checking {} for syntax errors...".format(self.module_name)
        ))

        # Check for syntax errors
        py_compile.compile(self.module_name + '.py')
        print(StatusMessage("No syntax errors found.", "success"))

        # Import module
        self.module = __import__(self.module_name)

        print()

        return True


    def run_custom_tests(self):
        raise NotImplementedError('No functionality tests implemented...')


    def run_style_tests(self):
        """Runs PEP8 compliance checking on student code"""
        print()
        print(HeaderMessage(
            "Checking {} for PEP8 compliance...".format(self.module_name)
        ))

        # Get the PEP8 results quietly
        style = pycodestyle.StyleGuide()

        with contextlib.redirect_stdout(io.StringIO()):
            result = style.check_files([self.module_name + '.py'])

        # Pretty print results
        result._deferred_print.sort()

        if not result._deferred_print:
            print(StatusMessage('No PEP8 violations found!', 'success'))
            return

        num_errors, num_warnings = 0, 0
        for line_number, offset, code, text, doc in result._deferred_print:
            if code[0] == 'E':
                num_errors += 1
            else:
                num_warnings += 1

        # Print out the counts
        err_msg = StatusMessage(
            f'{num_errors} error(s)',
            'fail'
        )

        warning_msg = StatusMessage(
            f'{num_warnings} warning(s)',
            'warning'
        )

        code_cmd = StatusMessage(
            f'$ pycodestyle {self.module_name}.py',
            'bold'
        )

        print(f"{err_msg} and {warning_msg} were found and suppressed.\n"
              f"Run {code_cmd} to see them.")

        print()


class BufferFalloffError(IndexError):
    """
    Error that is raised when someone tries to read an element from a buffer
    that doesn't exist.
    """
    pass


class RedirectStdin:
    """
    Serves as a dummy stdin object that you can write to and read from. 
    This object will then output from its self-stored data.
    """

    def __init__(self, to_output=None):
        # Build the list of things to output
        self.to_output = []
        if to_output:
            # Write initial buffer
            for e in to_output:
                self.write(e)

        # Initialize location to output from
        self.num_written = 0


    def write(self, s):
        """
        Writes s to self.to_output
        """
        # Kind of gross, but we need to make sure that it's readable
        if not isinstance(s, str):
            raise ValueError(
                f"Can't write a {type(s)} object to stdin (expected a string)."
            )

        self.to_output.append(s)


    def reset_buffer(self):
        """
        Resets the position from which the buffer is being read.
        """
        self.num_written = 0


    def clear(self):
        """
        Clears the buffer.
        """
        self.to_output = []
        self.num_written = 0


    def readline(self):
        """Returns the next element to_output"""
        if self.num_written >= len(self.to_output):
            raise BufferFalloffError(
                f"Cannot read element {self.num_written}. Has it been written?"
            )

        # Capture the element to output and return it
        out = self.to_output[self.num_written]
        self.num_written += 1

        # Print out to stdout for formatting
        print(out)
        
        return out


import setuptools

setuptools.setup(
    name='sp_autograder',
    packages=setuptools.find_packages(),
    version='0.2',
    license='MIT',
    description='A Python-backed customizable autograder.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    author='Parth Sarin',
    author_email='psarin@stanford.edu',
    url='https://github.com/stanfordpython/autograder',
    keywords=['EDUCATION', 'TESTING', 'AUTOGRADER'],
    install_requires=[
        'pycodestyle>=2.5.0',

    ],
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Topic :: Education',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ]
)



import sys
from .TestResponse import TestResponse
from autograder.printing import StatusMessage

def _dummy_setup_cleanup():
    pass

class BaseTest:
    def __init__(self,
                 student_obj=None,
                 solution_obj=None,
                 start_msg=None,
                 setup_fn=_dummy_setup_cleanup,
                 cleanup_fn=_dummy_setup_cleanup):
        """
        Initializes the BaseTest object which compares the student object to
        the solution object.

        Arguments
        ---------
        start_msg (str or None) -- The message to print at the beginning of the 
            test.
        setup_fn (function () -> None) -- The function that will be run before
            the test executes.
        cleanup_fn (function () -> None) -- The function that will be run after
            the test executes.
        """
        self.student_obj = student_obj
        self.solution_obj = solution_obj

        self.start_msg = start_msg
        if not start_msg:
            student_obj_name = student_obj.__name__
            self.start_msg = f'Testing {student_obj_name}...'

        self.start_msg = "{:68}".format(self.start_msg)

        self._setup_fn = setup_fn
        self._cleanup_fn = cleanup_fn


    def _handle_pass(self):
        """
        Prints out that the test passed.
        """
        diff = self.solution_response.diff(self.student_response)
        if diff and self.student_response.warning:
            print(StatusMessage('Warning.....', 'warning'))
            print(diff)
        else:
            print(StatusMessage('Test passed!', 'success'))


    @staticmethod
    def __handle_progressive_diff(diff):
        # Do Michael's progressive printing thing
        # 1. Find the error
        diff_lines = diff.split('\n')
        errors = filter(
            lambda idx_line: idx_line[1][0] == '+' or idx_line[1][0] == '-',
            enumerate(diff_lines)
        ) # does the line start with a + or a -?

        # 2. Print the header lines and capture the first error.
        print(next(errors)[1])
        print(next(errors)[1])
        fe_idx, fe_content = next(errors)

        # 3. Print the first error.
        print(fe_content)

        # 4. Ask up, down, or both
        msg = ("Type PRIOR to show all prior output, SUBSEQ for subsequent" 
               " output, or BOTH for a\nfull diff: ")
        resp = input(msg).upper()

        if resp == 'PRIOR':
            return '\n'.join(diff_lines[:fe_idx+1])
        elif resp == 'SUBSEQ':
            return '\n'.join(diff_lines[fe_idx-1:])
        else:
            return diff


    def _handle_fail(self):
        print(StatusMessage('Test failed!', 'fail'))

        diff = self.solution_response.diff(self.student_response)

        # Check if the progressive flag is set.
        is_progressive = '-p' in sys.argv or '--progressive' in sys.argv
        has_output_diff = self.solution_response.has_output_diff

        if is_progressive and has_output_diff:
            to_print = self.__handle_progressive_diff(diff)
        else:
            to_print = diff
        
        print(to_print)


    def _process_responses(self):
        """
        Processes the two responses and returns whether the test passed or
        failed.
        """
        output = self.solution_response == self.student_response

        if output:
            self._handle_pass()
        else:
            self._handle_fail()

        self._cleanup_fn()
        return output


    def _setup(self):
        """
        Sets up the test by printing the test start message and setting up by
        calling the provided setup function.
        """
        self._setup_fn()
        print(self.start_msg, end='')


    def run(self):
        """
        Runs the base test (just compares the two objects).
        """
        self._setup()

        self.solution_response = TestResponse(
            self.solution_obj, None, None, False, 'solution'
        )

        self.student_response = TestResponse(
            self.student_obj, None, None, False, 'student'
        )

        return self._process_responses()



from .ArgTest import ArgTest
from autograder.io_utils import RedirectStdin

class IOTest(ArgTest):
    def __init__(self,
                 student_obj=None,
                 solution_obj=None,
                 in_params=(),
                 *pos_args,
                 **key_args):
        # Initialize the ArgTest
        super().__init__(student_obj, solution_obj, 
                         *pos_args, **key_args)

        # Prepare the stdin buffer
        self.stdin_buffer = RedirectStdin(in_params)


    def run(self):
        """
        Runs the IO test (captures stdout, stderr, provides the given input
        to stdin, and runs both of the functions).
        """
        self._setup()

        # Run solution code and reset the buffer
        self.solution_response = self._captured_runner(
            self.solution_obj, self.args, self.kwargs, 'solution',
            f_stdin=self.stdin_buffer
        )
        self.stdin_buffer.reset_buffer()

        # Run student code and reset the buffer
        self.student_response = self._captured_runner(
            self.student_obj, self.args, self.kwargs, 'student',
            f_stdin=self.stdin_buffer
        )
        self.stdin_buffer.reset_buffer()

        return self._process_responses()


from .IOTest import IOTest
from autograder.io_utils import RedirectStdin

class FileIOTest(IOTest):
    def __init__(self,
                 filename,
                 student_obj=None,
                 solution_obj=None,
                 args=(),
                 kwargs={},
                 start_msg=None,
                 *pos_args,
                 **key_args):
        """
        Reads the lines from a file for the IOTest.
        """
        self.filename = filename
        
        with open(filename) as f:
            in_params = f.readlines()

        # Replace blank lines with \n
        in_params = [p if p else '\n' for p in in_params]

        # Initialize the IOTest
        super().__init__(student_obj, solution_obj, in_params, args, kwargs, 
                         start_msg, *pos_args, **key_args)


    def _serialize_args(self):
        """
        Builds a string representing the function call.
        """
        student_obj_name = self.student_obj.__name__
        
        args_lst = [repr(arg) for arg in self.args]
        kwargs_lst = [f'{k}={repr(v)}' for k, v in self.kwargs.items()]
        all_args = args_lst + kwargs_lst

        # Arguments and filename
        output = (f"{student_obj_name}({', '.join(all_args)}) "
                  f"(input from {self.filename})")

        # Try suppressing arguments
        if len(output) > 57:
            output = (f"{student_obj_name}(...several arguments...) "
                      f"(input from {self.filename})")

        # Otherwise suppress arguments message
        if len(output) > 57:
            output = f"{student_obj_name}(...) (input from {self.filename})"

        # Finally suppress filename
        if len(output) > 57:
            output = f"{student_obj_name}(...several arguments...)"


        return output